#undef MEMTRACE_ON
#include "memtracer.h"
#include "memtrace_int.h"
#include <zlib.h>
#include <ogc/lwp.h>
#include <ogc/mutex.h>
#include <string.h>

#ifdef USE_LOKI
#define LOKI_CLASS_LEVEL_THREADING
#include <Loki/SmallObj.h>
#endif


static FILE * trace_log = NULL;

static struct z_stream_s comp_stream;
static unsigned char out_buffer[1000];
static bool have_compress = false;

static mutex_t mut = LWP_MUTEX_NULL;


static void memtrace_at_exit();

extern "C" void initMemtrace(const char * file, int allow_compression)
{
	comp_stream.zalloc = NULL;
	comp_stream.zfree = NULL;
	comp_stream.opaque = NULL;
	comp_stream.next_in = NULL;
	comp_stream.avail_in = 0;
	comp_stream.next_out = out_buffer;
	comp_stream.avail_out = sizeof(out_buffer);
	
	if(allow_compression != 0)
	{
		if(deflateInit(&comp_stream, Z_DEFAULT_COMPRESSION) != Z_OK)
		{
			fprintf(stderr,"Do not have compression\n");
			have_compress = false;
		}
		else
		{
			have_compress = true;
		}
	}
	else
	{
		have_compress = false;
	}
	
	trace_log = NULL;
	trace_log = fopen(file,"wb");
	
	atexit(memtrace_at_exit);
	
	LWP_MutexInit(&mut, false);
		
	LWP_MutexLock(mut);
	
	LWP_MutexUnlock(mut);
	
}

extern "C" void closeMemtrace()
{
	memtrace_at_exit();
}

static void memtrace_at_exit()
{
	FILE* f = (FILE*)trace_log;
	
	LWP_MutexLock(mut);
	
	trace_log = NULL;
	
	LWP_MutexUnlock(mut);
	
	if(have_compress && f)
	{
		comp_stream.avail_in = 0;
		comp_stream.next_in = NULL;
		
		int err;
		for(;;)
		{
			if(comp_stream.avail_out == 0)
			{
				fwrite(out_buffer, sizeof(out_buffer), 1, f);
				fflush(f);
				comp_stream.next_out = out_buffer;
				comp_stream.avail_out = sizeof(out_buffer);
			}
			
			err = deflate(&comp_stream, Z_FINISH);
			
			if (err == Z_STREAM_END) break;
		};
				
		fwrite(out_buffer, sizeof(out_buffer)-comp_stream.avail_out, 1, f);
		fflush(f);
		
		deflateEnd(&comp_stream);
	}
	
	fflush(f);
	if(f) fclose(f);
}

static void write_comp(FILE *f, const void * data, size_t s)
{		
	comp_stream.next_in = (Bytef*)data;
	comp_stream.avail_in = s;
	
	do
	{
		if(comp_stream.avail_out == 0)
		{
			fwrite(out_buffer, sizeof(out_buffer), 1, f);
			fflush(f);
			comp_stream.next_out = out_buffer;
			comp_stream.avail_out = sizeof(out_buffer);
		}
		
		int err = deflate(&comp_stream, Z_NO_FLUSH);
		if(err < 0)
		{
			printf("Comp error %d\n", err);
		}
	} while(comp_stream.avail_in > 0);	
}

static void writedata(const void * data, size_t s)
{
	LWP_MutexLock(mut);
	
	FILE * f = trace_log;
	
	if(!f)
	{
		LWP_MutexUnlock(mut);
		return;
	}
	
	if(have_compress)
	{
		write_comp(f, data, s);
	}
	else
	{
		fwrite(data, s, 1, f);
	}
	
	LWP_MutexUnlock(mut);
}

#ifdef USE_LOKI
typedef Loki::AllocatorSingleton<> AllocatorSing;
#endif

extern "C" void * memtrace_malloc(const char * file, int line, size_t size )
{
	packet p;
	
#ifdef USE_LOKI
	static Loki::SmallObjAllocator &allocator = AllocatorSing::Instance();
	
	void * ptr = allocator.Allocate(size, false);
#else
	void * ptr = malloc(size);
#endif
	p.type = MEMTRACE_MALLOC;
	p.ptr = (int32_t)ptr;
	p.arg1 = size;
	p.arg2 = 0;
	p.line = line;
	p.strsize = strlen(file)+1;
	
	writedata(&p, sizeof(p));
	writedata(file, p.strsize);
	
	return ptr;
}

extern "C" void * memtrace_memalign(const char * file, int line, size_t blocksize, size_t bytes)
{
	packet p;
	
	void * ptr = memalign(blocksize, bytes);
	p.type = MEMTRACE_MEMALIGN;
	p.ptr = (int32_t)ptr;
	p.arg1 = blocksize;
	p.arg2 = bytes;
	p.line = line;
	p.strsize = strlen(file)+1;
	
	writedata(&p, sizeof(p));
	writedata(file, p.strsize);
	
	return ptr;
}

extern "C" void * memtrace_calloc  (const char * file, int line, size_t num, size_t size )
{
	packet p;

#ifdef USE_LOKI
	static Loki::SmallObjAllocator &allocator = AllocatorSing::Instance();
	
	void * ptr = allocator.Allocate(num*size,false);
	memset(ptr, 0, num*size);
#else
	void * ptr = calloc(num, size);
#endif
	
	p.type = MEMTRACE_CALLOC;
	p.ptr = (int32_t)ptr;
	p.arg1 = num;
	p.arg2 = size;
	p.line = line;
	p.strsize = strlen(file)+1;
	
	writedata(&p, sizeof(p));
	writedata(file, p.strsize);
	
	return ptr;
}

extern "C" void * memtrace_realloc (const char * file, int line, void * ptr, size_t size )
{
	packet p;

#ifdef USE_LOKI
	static Loki::SmallObjAllocator &allocator = AllocatorSing::Instance();
	void * ret_ptr = NULL;
	if(size > 0)
	{
		ret_ptr = allocator.Allocate(size,false);
		if(ptr != 0)
		{
			memcpy(ret_ptr, ptr, size);
		}	
	}
	allocator.Deallocate(ptr);
#else
	void * ret_ptr = realloc(ptr, size);
#endif

	p.type = MEMTRACE_REALLOC;
	p.ptr = (int32_t)ret_ptr;
	p.arg1 = (int32_t)ptr;
	p.arg2 = size;
	p.line = line;
	p.strsize = strlen(file)+1;
	
	writedata(&p, sizeof(p));
	writedata(file, p.strsize);
	
	return ret_ptr;
}

extern "C" void   memtrace_free(const char * file, int line, void * ptr )
{
	packet p;

#ifdef USE_LOKI
	static Loki::SmallObjAllocator &allocator = AllocatorSing::Instance();
	allocator.Deallocate(ptr);
#else
	free(ptr);
#endif
	
	p.type = MEMTRACE_FREE;
	p.ptr = 0;
	p.arg1 = (int32_t)ptr;
	p.arg2 = 0;
	p.line = line;
	p.strsize = strlen(file)+1;
	
	writedata(&p, sizeof(p));
	writedata(file, p.strsize);
}
